function [holdL, holdk, holdnumbclust, holdtheta, holdpred, pred] = sampler_NGG(data, theta, updatetheta, sampler, burnin, every, numbofits, predvalues)

% Calculating predictive distribution or not?
if ( nargin == 7 )
    preds = 0;
else
    preds = 1;
    pred = zeros(1, length(predvalues));
end

% Setting hyperparameters for the mixture model components
mu0 = mean(data);
sigmasq0 = 30^2;
sigmasqalpha = 1;
sigmasqbeta = 1;

% Set up total number of iterations
totalits = burnin + every * numbofits;

L = 10^(-2);

% Set up the holdtheta matrix to store samples of theta - empty if both
% updatelambda parameters are zero
if ( sum(updatetheta) == 0)
    holdtheta = [];
else
    holdtheta = zeros(2, numbofits);
end

% Set up matrices to store results
holdL = zeros(1, numbofits);
holdk = zeros(1, numbofits);
holdnumbclust = zeros(1, numbofits);
holdpred = zeros(1, numbofits);

n = length(data);

% Set up for calculating predictive or not
if (preds == 1)
    s = 1:n;
    sn = n + 1;
    kstar = n + 1;
else
    s = 1:n;
    kstar = n;
end

% Set jump sizes and parameters of components
J = zeros(1, kstar);
mu = mu0+sqrt(sigmasq0)*randn(1, kstar);
sigmasq = 1./gamrnd(sigmasqalpha, 1/sigmasqbeta, 1, kstar);

for i = 1:kstar
    J(i) = draw_NGG(L, theta(2));
end

% Set up slice variables
if ( preds == 1 )
    sumphi = gamrnd(n + 1, 1 / sum(J));
    if ( strcmp(sampler, 'slice1') == 1 )
        ustar = rand(1, n + 1) .* [J(s) J(sn)];
        L = min(ustar);
    else
        ustar = rand * min([J(s) J(sn)]);
        L = ustar;
    end
else
    sumphi = gamrnd(n, 1 / sum(J));
    if ( strcmp(sampler, 'slice1') == 1 )
        ustar = rand(1, n) .* J(s);
        L = min(ustar);
    else
        ustar = rand * min(J(s));
        L = ustar;
    end
end

phiaccept = 0;
phicount = 0;
aaccept = 0;
acount = 0;

for it = 1:totalits

    % Update allocation variables, s
    if ( strcmp(sampler, 'slice1') == 1 )
        for j = 1:n
            prob = (J>ustar(j)).*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));

            fprob = cumsum(prob);
            fprob = fprob / fprob(end);

            u = rand;
            s(j) = 1;
            while ( fprob(s(j)) < u )
                s(j) = s(j) + 1;
            end
        end
        if (preds == 1)
            prob = J>ustar(sn);

            fprob = cumsum(prob);
            fprob = fprob / fprob(end);

            u = rand;
            sn = 1;
            while ( fprob(sn) < u )
                sn = sn + 1;
            end
        end
    else
        for j = 1:n
            if ( preds == 0 )
                sminus = s([1:(j-1) (j+1):n]);
            else
                sminus = [sn s([1:(j-1) (j+1):n])];
            end
            [Jstar, findJstar] = min(J(sminus));
            findJstar = sminus(findJstar);
            prob = (J>Jstar).*J.*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));
            prob(findJstar) = J(findJstar)*exp(- 0.5 * (data(j) - mu(findJstar))^2 ./ sigmasq(findJstar) - 0.5 * log(sigmasq(findJstar)));
            prob = prob + (J<Jstar).*(J>ustar).*Jstar.*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));

            fprob = cumsum(prob);
            fprob = fprob / fprob(end);

            u = rand;
            s(j) = 1;
            while ( fprob(s(j)) < u )
                s(j) = s(j) + 1;
            end
        end

        if ( preds == 1)
            sminus = s;
            [Jstar, findJstar] = min(J(sminus));
            findJstar = sminus(findJstar);
            prob = (J>Jstar).*J.*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));
            prob(findJstar) = J(findJstar)*exp(- 0.5 * (data(j) - mu(findJstar))^2 ./ sigmasq(findJstar) - 0.5 * log(sigmasq(findJstar)));
            prob = prob + (J<Jstar).*(J>ustar).*Jstar.*exp(- 0.5 * (data(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));

            fprob = cumsum(prob);
            fprob = fprob / fprob(end);

            u = rand;
            sn = 1;
            while ( fprob(sn) < u )
                sn = sn + 1;
            end
        end
    end

    % Update the latent variable v (with the name sumphi in this code)
    newsumphi = sumphi * exp(0.5 * randn);

    int1 = theta(1) * quad(@(x) func2_general(x, newsumphi, @(x) func_NGG(x, theta)),0,L);
    int2 = theta(1) * quad(@(x) func2_general(x, sumphi,@(x) func_NGG(x, theta)),0,L);

    if ( preds == 1)
        newloglike = (n + 1) * log(newsumphi) - newsumphi * sum(J) - int1;
        loglike = (n + 1) * log(sumphi) - sumphi * sum(J) - int2;
    else
        newloglike = n * log(newsumphi) - newsumphi * sum(J) - int1;
        loglike = n * log(sumphi) - sumphi * sum(J) - int2;
    end
    logaccept = newloglike - loglike;

    accept = 1;
    if ( logaccept < 0 )
        accept = exp(logaccept);
    end

    phiaccept = phiaccept + accept;
    phicount = phicount + 1;

    if ( rand < accept )
        sumphi = newsumphi;
    end

    % Update jumps with observations allocated to them
    nj = zeros(1, length(J));
    for j = 1:n
        nj(s(j)) = nj(s(j)) + 1;
    end
    if ( preds == 1 )
        nj(sn) = nj(sn) + 1;
    end

    holds = [];
    counts = zeros(1, length(J));
    counter = 0;
    for j = 1:length(J)
        if ( nj(j) > 0 )
            holds = [holds j];
            counter = counter + 1;
            counts(j) = counter;
            J(j) = gamrnd(nj(j) - theta(2), 1 / (1 + sumphi));
        end
    end

    J = J(holds);
    mu = mu(holds);
    sigmasq = sigmasq(holds);
    s = counts(s);

    % Update parameters of components with observations allocated to them
    for j = 1:length(mu)
        meanstar = (sum(data(s==j)) / sigmasq(j) + mu0 / sigmasq0) / (sum(s==j) / sigmasq(j) + 1 / sigmasq0);
        varstar = 1 / (sum(s==j) / sigmasq(j) + 1 / sigmasq0);

        mu(j) = meanstar + sqrt(varstar) * randn;

        alphastar = sigmasqalpha + 0.5 * sum(s==j);
        betastar = sigmasqbeta + 0.5 * sum((data(s==j) - mu(j)).^2);

        sigmasq(j) = 1 / gamrnd(alphastar, 1 / betastar);
    end

    % Simulate slice variables
    if ( preds == 1 )
        sn = counts(sn);
        if ( strcmp(sampler, 'slice1') == 1 )
            ustar = rand(1, n + 1) .* [J(s) J(sn)];
            L = min(ustar);
        else
            ustar = rand * min([J(s) J(sn)]);
            L = ustar;
        end
    else
        if ( strcmp(sampler, 'slice1') == 1 )
            ustar = rand(1, n) .* J(s);
            L = min(ustar);
        else
            ustar = rand * min(J(s));
            L = ustar;
        end
    end
    
    % Simulation the jump process without observations allocated to them
    limit = int_general(L, @(x) func5_general(x, sumphi, @(x) func_NGG(x, theta)));

    kstar = min(10^7, poissrnd(theta(1) * limit));

    oldlengthJ = length(J);
    J = [J zeros(1, kstar)];
    mu = [mu mu0+sqrt(sigmasq0)*randn(1, kstar)];
    sigmasq = [sigmasq 1./gamrnd(sigmasqalpha, 1 / sigmasqbeta, 1, kstar)];

    Lstar = (1 + sumphi) * L;

    for j = (oldlengthJ+1):(oldlengthJ+kstar)
        F =  draw_NGG(Lstar, theta(2));
        J(j) = F / (1 + sumphi);
    end
    
    % Update M
    if ( updatetheta(1) == 1 )
        alphastar = 1 + length(J);
        int1 = quad(@(x) func2_general(x, sumphi, @(x) func_NGG(x, theta)),0,L);
        int2 = int_general(L, @(x) func_NGG(x, theta));
        betastar = 1 + int1 + int2;
        
        theta(1) = gamrnd(alphastar, 1 / betastar);
    end

    % Update a
    if ( updatetheta(2) == 1 )
        transa = log(theta(2)) - log(1 - theta(2));
        newtransa = transa + randn;
        newa = exp(newtransa) / (1 + exp(newtransa));
        newtheta  = theta;
        newtheta(2) = newa;
        
        newint = int_general(L, @(x) func3_general(x, sumphi, @(x) func_NGG(x, newtheta)));
        newloglike = theta(1) / newtheta(2) * (1 - (1 + sumphi)^newtheta(2)) - length(J) * gammaln(1 - newtheta(2)) - newtheta(2) * sum(log(J)) - theta(1) *  newint;
                
        int = int_general(L, @(x) func3_general(x, sumphi, @(x) func_NGG(x, theta)));
        loglike = theta(1) / theta(2) * (1 - (1 + sumphi)^theta(2)) - length(J) * gammaln(1 - theta(2)) - theta(2) * sum(log(J)) - theta(1) * int;
                 
        logaccept = newloglike - loglike +  log(1 / theta(2) + 1 / (1 - theta(2))) - log(1 / newa + 1 / (1 - newa));
        
        accept = 1;
        if ( logaccept < 0 )
            accept = exp(logaccept);
        end

        aaccept = aaccept + accept;
        acount = acount + 1;

        if ( rand < accept )
            theta = newtheta;
        end
    end

    % Record output of chain
    if ( (it > burnin) && (mod(it - burnin, every) == 0) )
        if ( sum(updatetheta) > 0 )
            holdtheta(:, (it - burnin) / every) = theta;
        end
        holdL((it - burnin) / every) = L;
        holdk((it - burnin) / every) = length(J);
        holdnumbclust((it - burnin) / every) = length(unique(s));

        % Rao-Blackwellised estimate of predictive density
        if ( preds == 1 )
            holdpred((it - burnin) / every) = mu(sn) + sqrt(sigmasq(sn)) * randn;

            if ( strcmp(sampler, 'slice1') )
                for j = 1:length(predvalues)
                    prob = (J>ustar(sn)).*exp(- 0.5 * (predvalues(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));

                    pred(j) = pred(j) + sum(prob)/sum(J>ustar(sn));
                end
            else
                sminus = s;
                [Jstar, findJstar] = min(J(sminus));
                findJstar = sminus(findJstar);
                for j = 1:length(predvalues)
                    prob = (J>Jstar).*J.*exp(- 0.5 * (predvalues(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));
                    prob(findJstar) = J(findJstar)*exp(- 0.5 * (predvalues(j) - mu(findJstar))^2 ./ sigmasq(findJstar) - 0.5 * log(sigmasq(findJstar)));
                    prob = prob + (J<Jstar).*(J>ustar).*Jstar.*exp(- 0.5 * (predvalues(j) - mu).^2 ./ sigmasq - 0.5 * log(sigmasq));
                    pred(j) = pred(j) + sum(prob) ./ sum(J(J>ustar));
                end
            end
        end
    end
    
    if ( mod(it, 100) == 0 )
        it
        phiaccept / phicount
        if ( updatetheta(2) == 1 )
            aaccept / acount
        end
        theta
        length(unique(s))
    end
end

if (preds == 1)
    pred = pred / numbofits;
else
    pred = [];
    holdpred = [];
end